/*
 *	Atom program to compute whole program path profiles
 *
 *	Purpose:
 *		Generate data to compare/contrast with statistical hotpath
 *		traces (a-la-Wiggins/Redstone)
 *		Further investigation into the path-based behavior of
 *		programs
 *		As a generalized hot path profiler.
 *
 *	Written by Rick Gorton July, 2000
 *	gorton@world.std.com
 *
 *	Copyright (C) 2000, Rick Gorton
 *	All rights reserved
 *
 *
 * NOTE:
 *	If you find this tool useful, the author is interested in knowing
 *	how you used it, and what kind of modifications you found useful.
 *
 */

#define	GENERATE_BLOCKMAP_FILE	1
#define	GENERATE_RAWPATH_FILE	1
#define	GENERATE_SUMMARY_FILE	1
#define	GENERATE_HOTPATH_FILE	1

#ifdef USE_INT_BLOCK_IDS
#define	NUM_ELTS_PER_LONG	2
#define LOG2_NUM_ELTS	1
#else
#define	NUM_ELTS_PER_LONG	4
#define LOG2_NUM_ELTS	2
#endif

#define	DEFAULT_NUMBER_HOTPATHS	256

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <signal.h>
#include "datatypes.h"

/*
 * Data declarations local to this file
 */
extern char	*Inst_Logfile_Name;

FILE		*RawPath_Logfile;
char		raw_path_filename[256];
FILE		*Summary_Logfile;
char		summary_filename[256];
FILE		*BlockMap_Logfile;
char		blockmap_filename[256];
FILE		*HotPath_Logfile;
char		hotpath_filename[256];

long	Total_Procs;
PROC_TABLE_TYPE		*dyn_proc_count;

long	Total_Static_Blocks;
long	Curr_Epoch;
BLOCK_TABLE_TYPE	*dyn_block_info;

NODE_TABLE_TYPE		*DynNodeCount;

/*
 * Whole program path profile variables
 */
long		Total_Paths;	// Overall number of path chains executed
long		Total_Unique_Paths;
long	VeryLongPaths;

#define INITIAL_PATH_BUF_SIZE	1024
int	extend_pbuf_index;
int	extend_pbuf[INITIAL_PATH_BUF_SIZE];
	// An initial buffer for Extend_Path to use.
	// By using a direct array, we eliminate some pointer chasing

#define HASH_INDEX_SIZE	256
	// HASH_INDEX_SIZE is used to define the storage arrays
	// for very fast lookup & comparison.  If this changes, the
	// HashVal function needs to change as well

#define PATH_LENGTH_INDEX_SIZE	256
SIMPLE_PATH_MULTI_ELT_TYPE *PathTable [PATH_LENGTH_INDEX_SIZE][HASH_INDEX_SIZE];

#define HOUSECLEANING_EPOCH_INTERVAL	0x400000
	// 0x400000 --> 4M
	// 0x1000000 --> 16M
	// Block interval at which to throw away the paths which aren't
	// important.  This is going to be very much heuristically driven

typedef struct {
	long				size;
	SIMPLE_PATH_MULTI_ELT_TYPE	*link;
	long				seqnum;	// For matching vs. rawpath contents
	} SORTED_PATHS_TYPE;

// Forward declarations
void Build_Path (long len);
void Print_Path (int seqnum, int length, SIMPLE_PATH_MULTI_ELT_TYPE *the_paths);
void Print_HotPath (int seqnum, int length, SIMPLE_PATH_MULTI_ELT_TYPE *the_paths);
void Discard_Low_Value_Paths (int is_final);
void DumpDynamicData();
void DumpDynamicDataAndExit();
/*
 * Initialization
 */

void DumpDynamicDataAndExit()
{
  fprintf (Inst_Logfile, "Signal detected - aborting at Epoch %ld\n", Curr_Epoch);
  DumpDynamicData();
  exit(1);
}

void DynamicApplicationInit()
{
  long	i, j;

  Total_Paths = 0;
  Total_Unique_Paths = 0;
  VeryLongPaths = 0x0;

  for (i = 0; i < INITIAL_PATH_BUF_SIZE; i++)
	extend_pbuf[i] = 0x0;
  extend_pbuf_index = -1;

  for (i = 0; i < PATH_LENGTH_INDEX_SIZE; i++)
    {
    for (j = 0; j < HASH_INDEX_SIZE; j++)
	PathTable [i][j] = NULL;
    }
  Curr_Epoch = 0;

	// And install some signal catchers so we can intentionally
	// force data to be dumped part-way through a run

  signal (SIGABRT, DumpDynamicDataAndExit);
  signal (SIGBUS, DumpDynamicDataAndExit);
  signal (SIGSEGV, DumpDynamicDataAndExit);

    if (GENERATE_RAWPATH_FILE == 1)
	{
	// For checking raw paths to see if duplicates are erroneously
	// being constructed
	sprintf (raw_path_filename, "%s.rawpath", Inst_Logfile_Name);
	RawPath_Logfile = fopen (raw_path_filename, "a+");
	fprintf (RawPath_Logfile, "Raw path data - Sequence Number: <block id list>\n");
	}

    if (GENERATE_SUMMARY_FILE == 1)
	{
	sprintf (summary_filename, "%s.summary", Inst_Logfile_Name);
	Summary_Logfile = fopen (summary_filename, "a+");
	fprintf (Summary_Logfile,
		"Program:Dynamic instructions:Static blocks:Dynamic blocks:Blocks covered:Paths executed:Unique paths\n");
	}

    if (GENERATE_BLOCKMAP_FILE == 1)
	{
	sprintf (blockmap_filename, "%s.blockmap", Inst_Logfile_Name);
	BlockMap_Logfile = fopen (blockmap_filename, "a+");
	fprintf (BlockMap_Logfile,
		"Program:blocks:blocks executed:blocks covered:Paths executed:Unique Paths\n");
	}

    if (GENERATE_HOTPATH_FILE == 1)
	{
	sprintf (hotpath_filename, "%s.hotpath", Inst_Logfile_Name);
	HotPath_Logfile = fopen (hotpath_filename, "a+");
	fprintf (HotPath_Logfile,
		"Hottest paths (by trip count) for %s\n", Inst_Logfile_Name);
	}

}

long GetInstrCountOfPath (int path_length, SIMPLE_PATH_MULTI_ELT_TYPE  *pchain)
{
    long	i, j;
    long	num_elts, num_path_insns;
    PATH_BLOCK_DATA	*raw_data;

    num_path_insns = 0;
    raw_data = pchain->dop.first_block;
    num_elts = (path_length + NUM_ELTS_PER_LONG - 1) >> LOG2_NUM_ELTS;
    if (num_elts == 1)
	raw_data = (PATH_BLOCK_DATA *) &pchain->dop.first_block;
    for (i = 0; i < num_elts; i++)
	{
	for (j = 0; j < NUM_ELTS_PER_LONG; j++)
	    {
	    long	id;

#ifdef USE_INT_BLOCK_IDS
	    id = raw_data[i].u.block_id_int[j];
#else
	    id = raw_data[i].u.block_id_ushort[j];
#endif
	    if (id > 0)
		{
		num_path_insns += dyn_block_info[id].Num_Insns;
		}
	    }
	}
    return num_path_insns;
}

/*
 * Print out results
 */
int Compare_InstrFreq_Paths (const void *v1, const void *v2)
{
  const SORTED_PATHS_TYPE	*el1, *el2;
  SIMPLE_PATH_MULTI_ELT_TYPE	*link1, *link2;
  long		len1, len2;

    el1 = v1;
    el2 = v2;
    link1 = (SIMPLE_PATH_MULTI_ELT_TYPE *) el1->link;
    link2 = (SIMPLE_PATH_MULTI_ELT_TYPE *) el2->link;

    len1 = link1->use_count * GetInstrCountOfPath (el1->size, link1);
    len2 = link2->use_count * GetInstrCountOfPath (el2->size, link2);
    return (len2 - len1);
}

int Compare_Paths (const void *v1, const void *v2)
{
  const SORTED_PATHS_TYPE	*el1, *el2;
  SIMPLE_PATH_MULTI_ELT_TYPE	*link1, *link2;

    el1 = v1;
    el2 = v2;
    link1 = (SIMPLE_PATH_MULTI_ELT_TYPE *) el1->link;
    link2 = (SIMPLE_PATH_MULTI_ELT_TYPE *) el2->link;

    return (link2->use_count - link1->use_count);
}

void DumpDynamicData()
{
  long	i, j, copy_index, path_print_count;
  SIMPLE_PATH_MULTI_ELT_TYPE	*path_walker;
  SORTED_PATHS_TYPE	*sorted_paths;
  char	*tmpstring;
  long	total_inst_count = 0;
  long	Total_Dynamic_Blocks = 0;
  long	coverage_block_count = 0;
  long	number_hotpaths_to_print;

    if (VeryLongPaths > 0)
	fprintf (Inst_Logfile,
		"A total of %ld very long paths of length > 255 blocks were executed\n",
		VeryLongPaths);

	/*
	 * Simple (Unique) Paths:
	 * Tidy up the (dangling) program exit path, and do other housekeeping
	 */
    Build_Path (extend_pbuf_index);
    Discard_Low_Value_Paths (TRUE);

	/*
	 * Figure out how many instructions and blocks got executed
	 */
    for (i = 0; i < Total_Static_Blocks; i++)
	{
	Total_Dynamic_Blocks += DynNodeCount[i].counter;
	total_inst_count += DynNodeCount[i].counter * dyn_block_info[i].Num_Insns;
	}

    fprintf (Inst_Logfile, "A Total of %ld instructions were executed\n",
		total_inst_count);
    fprintf (Inst_Logfile, "A Total of %ld blocks were executed\n",
		Total_Dynamic_Blocks);

    if (GENERATE_BLOCKMAP_FILE == 1)
	{
	/*
	 * Block Node counts
	 */
	fprintf (BlockMap_Logfile, "/* Block Profile data */\n");
	fprintf (BlockMap_Logfile, "/* Block id\tExecution Count\tAddress\t# of Insns */\n");
	}

    for (i = 0; i < Total_Static_Blocks; i++)
	{
	if (DynNodeCount[i].counter > 0)
	    {
	    coverage_block_count++;
	    if (GENERATE_BLOCKMAP_FILE == 1)
		fprintf (BlockMap_Logfile, "{ %ld,\t%ld,\t0x%lx,\t%ld }\n",
				i, DynNodeCount[i].counter,
				dyn_block_info[i].Block_Addr,
				dyn_block_info[i].Num_Insns);
	    }
	}

    fprintf (Inst_Logfile, "Covered %ld of %ld total blocks\n",
		coverage_block_count, Total_Static_Blocks);

    fprintf (Inst_Logfile, "\n*** Simple Program Paths ***\n");
    fprintf (Inst_Logfile, "A total of %ld Simple Program Paths were executed\n",
		Total_Paths);
    fprintf (Inst_Logfile, "A total of %ld Unique Simple Program Paths were found\n",
		Total_Unique_Paths);

    if (GENERATE_SUMMARY_FILE == 1)
	fprintf (Summary_Logfile, "%s:%ld:%ld:%ld:%ld:%ld:%ld\n",
			Inst_Logfile_Name,
			total_inst_count,
			Total_Static_Blocks,
			Total_Dynamic_Blocks,
			coverage_block_count,
			Total_Paths,
			Total_Unique_Paths);

    if ((DUMP_OUTPUT_MODE & DUMP_PATH_SOURCE_INFO) == DUMP_PATH_SOURCE_INFO)
	fprintf (Inst_Logfile,
		"*** NOTE *** Source file and line numbers of (null) and 0 respectively means\n"
		"*** NOTE *** that the information is NOT available in the symbol table.\n");

	/*
	 * Figure out the path vs. execution distributions and print that info out
	 * to the logfile.
	 */

    fprintf (Inst_Logfile, "\nPath length distribution information\n"
			   "Path Length:Number of Unique:Aggregate executions:Aggregate\n"
			   "(N blocks) :Paths (Length N):of Paths (Length N) :Instructions\n");

	/*
	 * Traverse the paths to set up for sorting them by frequency.
	 * Also print out path distribution data since we are already
	 * traversing the table.
	 */
    sorted_paths = (SORTED_PATHS_TYPE *) calloc (sizeof(SORTED_PATHS_TYPE),
								Total_Unique_Paths);
    copy_index = 0;
    for (i = 0; i < PATH_LENGTH_INDEX_SIZE; i++)
	{
	long	LenDist_Unique, LenDist_Aggregate, LenDist_Instructions;

	LenDist_Unique = 0;
	LenDist_Aggregate = 0;
	LenDist_Instructions = 0;
	for (j = 0; j < HASH_INDEX_SIZE; j++)
	    {
	    SIMPLE_PATH_MULTI_ELT_TYPE	*pchain;

	    pchain = PathTable [i][j];
	    while (pchain != NULL)
		{
		LenDist_Unique++;
		LenDist_Aggregate += pchain->use_count;
		LenDist_Instructions += (pchain->use_count * GetInstrCountOfPath (i, pchain));
		sorted_paths[copy_index].size = i;
		sorted_paths[copy_index].link = pchain;
		sorted_paths[copy_index].seqnum = copy_index;
		copy_index++;
		pchain = pchain->next_path;
		}
	    }
	if (LenDist_Unique > 0)
	    fprintf (Inst_Logfile, "%ld:%ld:%ld:%ld\n",
			i, LenDist_Unique, LenDist_Aggregate, LenDist_Instructions);
	}

    qsort (sorted_paths, Total_Unique_Paths, sizeof(SORTED_PATHS_TYPE), Compare_Paths);

	/*
	 * And now print them all out
	 */
    fprintf (Inst_Logfile, "\nSimple unique path data (sorted by path execution count):\n"
	   "(Match the sequence number to the rawpath sequence number for block details)\n"
	   "Sequence Number:Execution Count:Block Length:Instruction Length:Creation Epoch:Last use Epoch\n");
    for (i = 0; i < Total_Unique_Paths; i++)
	{
	Print_Path (sorted_paths[i].seqnum, sorted_paths[i].size, sorted_paths[i].link);
	}

    tmpstring = getenv ("NUMBER_HOTPATHS");
    if (tmpstring == NULL)
	number_hotpaths_to_print = DEFAULT_NUMBER_HOTPATHS;
    else
	number_hotpaths_to_print = atol (tmpstring);

	/*
	 * And print out the hottest paths by path execution frequency
	 * to the hotpath file.
	 */
    if (Total_Unique_Paths < number_hotpaths_to_print)
	number_hotpaths_to_print = Total_Unique_Paths;

    fprintf (HotPath_Logfile, "\nSimple unique path data (sorted by path execution count):\n"
	   "(Match the sequence number to the rawpath sequence number for block details)\n"
	   "Sequence Number:Execution Count:Block Length:Instruction Length:Creation Epoch:Last use Epoch\n");
    if (GENERATE_HOTPATH_FILE == 1)
	{
	for (i = 0; i < number_hotpaths_to_print; i++)
	    {
	    Print_HotPath (sorted_paths[i].seqnum, sorted_paths[i].size, sorted_paths[i].link);
	    }

	qsort (sorted_paths, Total_Unique_Paths, sizeof(SORTED_PATHS_TYPE), Compare_InstrFreq_Paths);

	fprintf (HotPath_Logfile, "\nSimple unique path data (sorted by INSTRUCTION execution count):\n"
	   "(Match the sequence number to the rawpath sequence number for block details)\n"
	   "Sequence Number:Execution Count:Block Length:Instruction Length:Creation Epoch:Last use Epoch\n");
	for (i = 0; i < number_hotpaths_to_print; i++)
	    {
	    Print_HotPath (sorted_paths[i].seqnum, sorted_paths[i].size, sorted_paths[i].link);
	    }
	}

    if (GENERATE_RAWPATH_FILE == 1)
	{
	fflush (RawPath_Logfile);
	fclose (RawPath_Logfile);
	}
    if (GENERATE_SUMMARY_FILE == 1)
	{
	fflush (Summary_Logfile);
	fclose (Summary_Logfile);
	}
    if (GENERATE_BLOCKMAP_FILE == 1)
	{
	fprintf (BlockMap_Logfile, "/* End of block map */\n");
	fflush (BlockMap_Logfile);
	fclose (BlockMap_Logfile);
	}
    if (GENERATE_HOTPATH_FILE == 1)
	{
	fflush (HotPath_Logfile);
	fclose (HotPath_Logfile);
	}
    fprintf (Inst_Logfile, "End of dynamic data\n");
    fflush (Inst_Logfile);
	// fclose of Inst_Logfile done elsewhere...
    return;
}

void Print_Path (int seqnum, int path_length, SIMPLE_PATH_MULTI_ELT_TYPE *pchain)
{
    long	i, j;
    long	num_elts, num_path_insns;
    PATH_BLOCK_DATA	*raw_data;

    num_path_insns = 0;
    raw_data = pchain->dop.first_block;
    num_elts = (path_length + NUM_ELTS_PER_LONG - 1) >> LOG2_NUM_ELTS;
    if (num_elts == 1)
	raw_data = (PATH_BLOCK_DATA *) &pchain->dop.first_block;
    for (i = 0; i < num_elts; i++)
	{
	for (j = 0; j < NUM_ELTS_PER_LONG; j++)
	    {
	    long	id;

#ifdef USE_INT_BLOCK_IDS
	    id = raw_data[i].u.block_id_int[j];
#else
	    id = raw_data[i].u.block_id_ushort[j];
#endif
	    if (id > 0)
		{
		num_path_insns += dyn_block_info[id].Num_Insns;
		}
	    }
	}

    fprintf (Inst_Logfile,
		"%ld:%ld:%ld:%ld:%ld:%ld\n",
		seqnum, pchain->use_count, path_length, num_path_insns,
		pchain->creation_epoch - path_length,
		pchain->last_used_epoch - path_length);

	// Print out the path as ID numbers
    if (GENERATE_RAWPATH_FILE == 1)
	fprintf (RawPath_Logfile, "%ld:", seqnum);
    for (i = 0; i < num_elts; i++)
	{
	for (j = 0; j < NUM_ELTS_PER_LONG; j++)
	    {
	    long	id;

#ifdef USE_INT_BLOCK_IDS
	    id = raw_data[i].u.block_id_int[j];
#else
	    id = raw_data[i].u.block_id_ushort[j];
#endif
	    if (id > 0)
		{
		if (GENERATE_RAWPATH_FILE == 1)
			fprintf (RawPath_Logfile, "\t%ld", id);
		}
	    }	// j loop
	}	// i loop

    if (GENERATE_RAWPATH_FILE == 1)
	{
	fprintf (RawPath_Logfile, "\n");
	fflush (RawPath_Logfile);
	}
    fflush (Inst_Logfile);
}

void Print_HotPath (int seqnum, int path_length, SIMPLE_PATH_MULTI_ELT_TYPE *pchain)
{
    long	i, j;
    long	num_elts, num_path_insns;
    PATH_BLOCK_DATA	*raw_data;

    num_path_insns = GetInstrCountOfPath (path_length, pchain);
    raw_data = pchain->dop.first_block;
    num_elts = (path_length + NUM_ELTS_PER_LONG - 1) >> LOG2_NUM_ELTS;
    if (num_elts == 1)
	raw_data = (PATH_BLOCK_DATA *) &pchain->dop.first_block;

    fprintf (HotPath_Logfile,
		"%ld:%ld:%ld:%ld:%ld:%ld\n",
		seqnum, pchain->use_count, path_length,
		num_path_insns, pchain->creation_epoch - path_length,
		pchain->last_used_epoch - path_length);

	// Print out the path as source file and line or addresses if the line
	// number information isn't there
	for (i = 0; i < num_elts; i++)
	    {
	    for (j = 0; j < NUM_ELTS_PER_LONG; j++)
		{
		long	id;

#ifdef USE_INT_BLOCK_IDS
		id = raw_data[i].u.block_id_int[j];
#else
		id = raw_data[i].u.block_id_ushort[j];
#endif
		if (id > 0)
		    {
		    long	proc_id;

		    proc_id = dyn_block_info[id].ProcId;
		    if (dyn_proc_count[proc_id].ProcFileName[0] != 0x0)
		        fprintf (HotPath_Logfile, "\t%s: ",
					dyn_proc_count[proc_id].ProcFileName);
		    else
		        fprintf (HotPath_Logfile, "\t%s() ",
					dyn_proc_count[proc_id].ProcName);

		    if (dyn_block_info[id].StartLineNum == 0)
			fprintf (HotPath_Logfile, " @ 0x%lx\n", dyn_block_info[id].Block_Addr);
		    else
			{
			if (dyn_block_info[id].StartLineNum == dyn_block_info[id].EndLineNum)
			    fprintf (HotPath_Logfile, " Line %ld\n",
							dyn_block_info[id].StartLineNum);
			else
			    fprintf (HotPath_Logfile, " Lines %ld - %ld\n",
							dyn_block_info[id].StartLineNum,
							dyn_block_info[id].EndLineNum);
			}
		    }	// id > 0
		}	// j loop
	    }	// i loop

    if (GENERATE_HOTPATH_FILE == 1)
	{
	fprintf (HotPath_Logfile, "\n");
	fflush (HotPath_Logfile);
	}
    fflush (HotPath_Logfile);
}

/*
 * Procedure data
 */

void DynAllocateStorage (long Proc_Elts, long Block_Elts)
{
  Total_Procs = Proc_Elts;
  dyn_proc_count = (PROC_TABLE_TYPE *) calloc (sizeof(PROC_TABLE_TYPE), Proc_Elts);

  Total_Static_Blocks = Block_Elts;
  dyn_block_info = (BLOCK_TABLE_TYPE *) calloc (sizeof(BLOCK_TABLE_TYPE), Block_Elts);

  DynNodeCount = (NODE_TABLE_TYPE *) calloc (sizeof (NODE_TABLE_TYPE), Block_Elts);

}

void Dyn_ProcCountInit (long PSeq, char *Pname, char *PFilename, long Addr)
{
    strncat (dyn_proc_count[PSeq].ProcName, Pname, PROC_NAME_SIZE);
    strncat (dyn_proc_count[PSeq].ProcFileName, PFilename, PROC_NAME_SIZE);
    dyn_proc_count[PSeq].StartAddr = Addr;
}

void BbInfo (long this_BlockNum, long this_ProcNum,
		long this_BlockSize, long this_BlockAddr,
		long this_StartLine, long this_EndLine)
{
	/*
	 * Called once before the program runs, used solely for
	 * cross-correlation of stuff at data output time.
	 */
  dyn_block_info[this_BlockNum].ProcId = this_ProcNum;
  dyn_block_info[this_BlockNum].Num_Insns = this_BlockSize;
  dyn_block_info[this_BlockNum].Block_Addr = this_BlockAddr;
  dyn_block_info[this_BlockNum].StartLineNum = this_StartLine;
  dyn_block_info[this_BlockNum].EndLineNum = this_EndLine;
}

	/*
	 * Cleaning heuristics
	 */

int Is_Low_Value_Path (SIMPLE_PATH_MULTI_ELT_TYPE *path_p, int plen, int is_final)
{

	/*
	 * Heuristics are:
	 *	'old' paths with use count == 1
	 *	Define old as a path which hasn't been used for more than cleaning interval.
	 *	So for a cleaning interval of 16e6 blocks, this will be clean out
	 *	something like XXX blocks each time.
	 *	Retire 'very old' paths (very old being interval * 2)
	 *	On the final cleanup, get rid of _all_ paths with counts of 1
	 *
	 *	Observation: for 099.go (cpu95), this results in a fairly consistent
	 *	removal of about
	 */

    if ((path_p->use_count == 1) &&
	((Curr_Epoch - path_p->last_used_epoch > HOUSECLEANING_EPOCH_INTERVAL)
	 || (is_final)))
	return TRUE;

    if (Curr_Epoch - path_p->last_used_epoch > (HOUSECLEANING_EPOCH_INTERVAL << 1))
	{
	fprintf (Inst_Logfile, "Retiring ");
	fprintf (Inst_Logfile,
		"SeqNum\tExecution count\tBlock Length\tInstruction Length\n");
	Print_Path (0, plen, path_p);
	return TRUE;
	}

    return FALSE;
}

void Discard_Low_Value_Paths(int is_final)
{
	/*
	 * This walks PathTable and throws out paths which are
	 * determined to be of little value.
	 */
  long	i, j;
  long	total_unique;
  long	low_count_deletions;

return;
    total_unique = 0;
    low_count_deletions = 0;
    for (i = 0; i < PATH_LENGTH_INDEX_SIZE; i++)
	{
	long	num_elts;

	num_elts = (i + NUM_ELTS_PER_LONG - 1) >> LOG2_NUM_ELTS;
	for (j = 0; j < HASH_INDEX_SIZE; j++)
	    {
	    SIMPLE_PATH_MULTI_ELT_TYPE	*pchain_p, *pred_p;

	    pred_p = PathTable[i][j];
	    if (pred_p != NULL)
		{
		// Singly linked list (and no, we really can't make it doubly linked)
		// due to storage constraints.  Instead, we'll delete low-value follow-on
		// paths first, and then check the roots later.
		total_unique++;
		pchain_p = pred_p->next_path;
		while (pchain_p != NULL)
		    {
		    total_unique++;
		    if (Is_Low_Value_Path (pchain_p, i, is_final))
			{
			low_count_deletions++;
			if (num_elts > 1)
			    free (pchain_p->dop.first_block);
			pred_p->next_path = pchain_p->next_path;
			free (pchain_p);
			pchain_p = pred_p;
			}
		    pchain_p = pchain_p->next_path;
		    }
		// Now check the root: PathTable[i][j]
		// We know for a fact that PathTable[i][j] != NULL...
		if (Is_Low_Value_Path (PathTable[i][j], i, is_final))
		    {
		    SIMPLE_PATH_MULTI_ELT_TYPE	*next_p;

		    low_count_deletions++;
		    if (num_elts > 1)
			free (PathTable[i][j]->dop.first_block);
		    next_p = PathTable[i][j]->next_path;
		    free (PathTable[i][j]);
		    PathTable[i][j] = next_p;
		    }
		}	// If PathTable[i][j] had entries
	    }	// j loop
	}	// i loop
    if (low_count_deletions > 0)
	{
	fprintf (Inst_Logfile, "Epoch: %ld - Discarded %ld low use of %ld total paths\n",
		Curr_Epoch, low_count_deletions, total_unique);
	fflush (Inst_Logfile);
	}
}
	/*
	 * Various path collection code routines
	 */

static int HashVal (int a, int b)
{
	// This is dependent upon the number of bits in HASH_INDEX_SIZE
	// Consider the case where the length is one - in this case, the property
	// a == b holds true.  In order to get some dispersion/distribution, we
	// need to look at different bits of a and b.
  return (a & 0xf0) | (b & 0xf);
}

void Build_Path (long pbuf_idx)
{
  int	hash_val;
  int	length_index;
  int	num_elts;
  long	i;
  SIMPLE_PATH_MULTI_ELT_TYPE	*existing_path;

    Total_Paths++;
	/*
	 * HACK: for very long paths, truncate things.  This could be
	 * improved by handling very long paths as a special case in
	 * the future, but it is presumably not particularly important,
	 * as they seldom occur, and so far, been found _only_ in gcc
	 * so far.
	 */

    if (pbuf_idx >= PATH_LENGTH_INDEX_SIZE - 2)
	{
	fprintf (Inst_Logfile,
		"Ignoring a very long path: length == %ld\tepoch: %ld\n",
		pbuf_idx, Curr_Epoch - pbuf_idx);
	VeryLongPaths++;
	return;
	}

	/*
	 * A completed path exists as extend_pbuf[0:pbuf_idx]
	 * Look for an existing match.
	 */

    hash_val = HashVal (extend_pbuf[0], extend_pbuf[pbuf_idx]);
    length_index = pbuf_idx + 1;

	// Compute number of elements to search, and zero out the next extend_pbuf
	// item(s) so our search won't have to collide with from previous path data.

    num_elts = (length_index + NUM_ELTS_PER_LONG - 1) >> LOG2_NUM_ELTS;
#ifdef USE_INT_BLOCK_IDS
    extend_pbuf [pbuf_idx + 1] = 0;
#else
    extend_pbuf [pbuf_idx + 1] = 0;
    extend_pbuf [pbuf_idx + 2] = 0;		// clear other previous entries
    extend_pbuf [pbuf_idx + 3] = 0;		// for this compression scheme
#endif

    existing_path = PathTable [length_index][hash_val];
    while (existing_path != NULL)
	{
	int	found;
	PATH_BLOCK_DATA	*raw_data;

	found = 1;
	raw_data = existing_path->dop.first_block;
	if (num_elts == 1)
	    raw_data = (PATH_BLOCK_DATA *) &existing_path->dop.first_block;
	for (i = 0; i < num_elts; i++)
	    {
#ifdef USE_INT_BLOCK_IDS
	    if ((raw_data[i].u.block_id_int[0] != extend_pbuf[2*i])
		|| (raw_data[i].u.block_id_int[1] != extend_pbuf[2*i+1]))
#else
	    if ((raw_data[i].u.block_id_ushort[0] != extend_pbuf[4*i])
		|| (raw_data[i].u.block_id_ushort[1] != extend_pbuf[4*i+1])
		|| (raw_data[i].u.block_id_ushort[2] != extend_pbuf[4*i+2])
		|| (raw_data[i].u.block_id_ushort[3] != extend_pbuf[4*i+3]))
#endif
		{
		found = 0;
		break;
		}
	    }

	if (found == 1)
	    {
		/*
		 * update the path info and we're done
		 */
	    existing_path->use_count++;
	    existing_path->last_used_epoch = Curr_Epoch;
	    return;
	    }

	existing_path = existing_path->next_path;
	}

    if (existing_path == NULL)
	{
	/*
	 * Construct a new path element, and insert at the front of the list
	 * If there is some temporal locality of paths (this whole experiment is
	 * predicated upon this assumption), this provides much faster searching
	 * than appending it.
	 */
	SIMPLE_PATH_MULTI_ELT_TYPE	*new_path;
	PATH_BLOCK_DATA		*raw_data;

	new_path = malloc (sizeof (SIMPLE_PATH_MULTI_ELT_TYPE));
	new_path->creation_epoch = Curr_Epoch;
	new_path->last_used_epoch = Curr_Epoch;
	new_path->use_count = 1;
	if (num_elts > 1)
	    {
	    new_path->dop.first_block = (PATH_BLOCK_DATA *) calloc (num_elts, sizeof(PATH_BLOCK_DATA));
	    raw_data = new_path->dop.first_block;
	    }
	else
	    raw_data = (PATH_BLOCK_DATA *) &new_path->dop.first_block;

	for (i = 0; i < num_elts; i++)
	    {
#ifdef USE_INT_BLOCK_IDS
	    raw_data[i].u.block_id_int[0] = extend_pbuf[2*i];
	    raw_data[i].u.block_id_int[1] = extend_pbuf[2*i+1];
#else
	    raw_data[i].u.block_id_ushort[0] = extend_pbuf[4*i];
	    raw_data[i].u.block_id_ushort[1] = extend_pbuf[4*i+1];
	    raw_data[i].u.block_id_ushort[2] = extend_pbuf[4*i+2];
	    raw_data[i].u.block_id_ushort[3] = extend_pbuf[4*i+3];
#endif
	    }

	new_path->next_path = PathTable [length_index][hash_val];
	PathTable [length_index][hash_val] = new_path;
	// fprintf (Inst_Logfile, "Building ");
	// fprintf (Inst_Logfile,
	//	"SeqNum\tExecution count\tBlock Length\tInstruction Length\n",
	// Print_Path (0, length_index, new_path);
	Total_Unique_Paths++;
	return;
	}

    fprintf (Inst_Logfile, "Build_Path failure: existing_path != NULL\n");
    fflush (Inst_Logfile);
    exit (1);
}

void Extend_Path (long this_BlockNum)
{
  long			found;
  long			i;

    Curr_Epoch++;
    DynNodeCount[this_BlockNum].counter++;

	// Look for a cycle in the temporary storage array
	// Do this in an array so we don't have to continuously do
	// pointer chasing and data loads off of those pointers.

    for (i = 0; i <= extend_pbuf_index; i++)
	{
	if (extend_pbuf[i] == this_BlockNum)
	    {
	    Build_Path (extend_pbuf_index);
	    extend_pbuf_index = -1;
	    }
	}

    extend_pbuf_index++;
    extend_pbuf [extend_pbuf_index] = this_BlockNum;
    if (extend_pbuf_index == INITIAL_PATH_BUF_SIZE)
	{
	// Huge path.  Flag it and continue
	fprintf (Inst_Logfile, "INITIAL_PATH_BUF_SIZE limit reached - resetting\n");
	extend_pbuf_index = -1;
	}

return;		// Not necessary to prune (at least not now)
    if ((Curr_Epoch % HOUSECLEANING_EPOCH_INTERVAL) == 0)
	{
	Discard_Low_Value_Paths (FALSE);
	}
}
