
/*
 *	Atom program to compute whole program path profiles
 *	Purpose:
 *		Generate data to compare/contrast with Wiggins/Redstone
 *		generated traces.
 *
 *	Copyright (C) 2000, Rick Gorton
 *
 *	Written by Rick Gorton, July, 2000
 *	gorton@world.std.com
 *	All rights reserved
 *
 * NOTE:
 *	If you find this tool useful, the author is interested in knowing
 *	how you used it, and what kind of modifications you found useful.
 *
 *
 * Technical notes and assumptions:
 *	This tool derived from atom_wholepath, but does a lot more compression,
 *	and is (hopefully) much faster than atom_wholepath.
 *
 *	Storage issues:
 *	) Memory consumption is the limiting factor, as opposed to compute time spent
 *	constructing path information.
 *
 *	) From looking at SPEC CPU95, I have observed that the _longest_ cyclic path
 *	happens in 126.gcc, and is 314 blocks in length.  The vast majority of path
 *	lengths in other programs seems to have lengths less than 20 blocks long.
 *
 *	) The number of basic blocks in these programs is usually < 30K, with gcc
 *	being the largest at 80K blocks when statically linked.  30K can be
 *	represented in 15 bits, while it takes 17 bits to handle 80K blocks.
 *	If someone wants to use this tool on an exceptionally large program, we
 *	need to leave a few extra bits for ids.  Using spike to remove unreachable
 *	code could reduce this.
 *
 *	) A simple block coverage tool shows that even for gcc (the biggest and most
 *	complex CPU95 program) the number of blocks which are actually executed is a
 *	few less than 32K.  So a dynamic mapping of physical block ids to virtual block
 *	id's could enable encoding of block ids using short ints (barely).  Using
 *	unsigned short ints gets us up to 64K of actual blocks.
 *	Doing this kind of conversion will entail extra computations, so don't do
 *	this compression unless the input program actually has > 64K basic blocks.
 *	The 64K dynamic block limit is a further hard limit, which ought to be
 *	relatively simple to #ifdef code for (at build time).
 *	Later note: the overhead is non-trivial, so this isn't being done.  Just
 *	go straight to using integer block ids.
 *
 *	) pointers and longs on Tru64 on Alpha are 8 bytes long, with ints being 4 bytes.
 *
 *	) If we dynamically allocate storage using *alloc(), assume that our minimum
 *	overhead is roughly (pointer to storage + a long for sizeof(storage area)).
 *	Which means that dynamic allocation is not going to be a win unless we are 
 *	allocating three or more longs.
 *	The reality of the trade-off is that the computation gets much more complex
 *	for the first part of paths, and wastes space for short paths (< 3 longs)
 *	so this isn't being done for now.
 *
 */

#include <stdio.h>
#include <cmplrs/atom.inst.h>
#include "datatypes.h"
#include "declare.h"

static long	Total_Num_Blocks;

void InstrumentInit(int iargc, char **iargv)
{
  int	i,j;
  char	namebuffer[256];

  if (iargc > 1)
    strcpy(namebuffer, iargv[1]);
  else
    strcpy(namebuffer, LOGFILE_NAME_);
  Inst_Logfile = fopen(namebuffer, "a+");
  printf("Logfile will be '%s'\n", namebuffer);
  fprintf(Inst_Logfile, "Logfile:\t%s\n", namebuffer);
  fprintf (Inst_Logfile, "*** Static data begin ***\n");

  ProcSequenceNum = 1;
  Total_Num_Blocks = 0;
  BlockSequenceNum = 1;
	/*
	 * Because of analysis-time constraints, this cannot be 0.
	 * which means we need to have loops using BlockSequenceNum
	 * start at 1 instead of 0.
	 * For consistency, do the same for ProcSequenceNum.
	 */

  AddCallProto ("BbInfo(long, long, long, long, long, long)");
  AddCallProto ("Extend_Path(long)");

  AddCallProto("DynamicInit(char *)");
  AddCallProgram(ProgramBefore, "DynamicInit", namebuffer);
}

Instrument(int argc, char **argv, Obj *obj)
{
  Proc *p; Block *b;  Inst *this_inst;
  RegvType	RegA, RegB, RegC;
  long		this_PC;
  enum Opcodes	curr_opcode;

  fprintf(Inst_Logfile, "Object:\t%s\n", GetObjName(obj));
  fprintf(Inst_Logfile, "Number of Procedures:\t%d\n", GetObjInfo(obj, ObjNumberProcs));

	/*
	 * Do this 'quick' scan to see if the number of blocks we are instrumenting
	 * requires that we do physical --> dynamic block mapping.
	 */
  for (p = GetFirstObjProc(obj); p != NULL; p = GetNextProc(p))
    {
    for (b = GetFirstBlock (p); b != NULL; b = GetNextBlock (b))
	Total_Num_Blocks++;
    }

#ifndef USE_INT_BLOCK_IDS
  if (Total_Num_Blocks >= USHRT_MAX)
    {
    fprintf (Inst_Logfile, "*** NOTE *** more than 64K blocks - rebuild using integer block ids\n");
    fprintf (Inst_Logfile, "*** NOTE *** by enabling #define USE_INT_BLOCK_IDS in datatypes.h\n");
    fprintf (Inst_Logfile, "*** NOTE *** (or linking against inst_intid.o)\n");
    fprintf (Inst_Logfile, "*** NOTE *** This will significantly increase memory consumption.\n");
    printf ("*** NOTE *** more than 64K blocks - rebuild using integer block ids\n");
    printf ("*** NOTE *** by enabling #define USE_INT_BLOCK_IDS in datatypes.h\n");
    printf ("*** NOTE *** (or linking against inst_intid.o)\n");
    printf ("*** NOTE *** This will significantly increase memory consumption.\n");
    exit (1);
    }
#endif

  for (p = GetFirstObjProc(obj); p != NULL; p = GetNextProc(p))
    {
	/*
	 * Build ProcTable up so we can pass its contents to the runtime
	 * _after_ we allocate space for it.
	 */
    if (ProcName(p) == NULL)
	strncat (&ProcTable[ProcSequenceNum].ProcName[0], "(null)", PROC_NAME_SIZE);
    else
	strncat (&ProcTable[ProcSequenceNum].ProcName[0], ProcName(p), PROC_NAME_SIZE);
    if (ProcFileName(p) == NULL)
	// .anal file triggers on a 0x0 as the first character to
	// mean unknown procedure name.
	ProcTable[ProcSequenceNum].ProcFileName[0] = 0x0;
    else
	strncat (&ProcTable[ProcSequenceNum].ProcFileName[0], ProcFileName(p), PROC_NAME_SIZE);
    ProcTable[ProcSequenceNum].StartAddr = ProcPC(p);

/*
OBSOLETE - Takes a lot of space up, and isn't particularly useful

    fprintf(Inst_Logfile, "%s (%ld)\t@ 0x%lx Block IDs %ld",
			ProcTable[ProcSequenceNum].ProcName, ProcSequenceNum,
			ProcPC(p), BlockSequenceNum);
*/

    for (b = GetFirstBlock (p); b != NULL; b = GetNextBlock (b))
	{
	Inst	*first_i, *last_i;

	first_i = GetFirstInst (b);
	last_i = GetLastInst (b);
	/*
	 * Build BlockTable and pass it's contents after allocating space for it.
	 */
	BlockTable[BlockSequenceNum].ProcId = ProcSequenceNum;
	BlockTable[BlockSequenceNum].Num_Insns = GetBlockInfo (b, BlockNumberInsts);
	BlockTable[BlockSequenceNum].Block_Addr = BlockPC (b);
	BlockTable[BlockSequenceNum].StartLineNum = InstLineNo (first_i);
	BlockTable[BlockSequenceNum].EndLineNum = InstLineNo (last_i);

	/*
	 * Record our execution of this block.
	 */
	AddCallBlock (b, BlockBefore, "Extend_Path", BlockSequenceNum);

	BlockSequenceNum++;
	if (BlockSequenceNum > BLOCK_TABLE_TYPE_SIZE)
	    {
	    fprintf (stderr, "BlockSequenceNum (%ld) > BLOCK_TABLE_TYPE_SIZE (%ld)\n",
		BlockSequenceNum, BLOCK_TABLE_TYPE_SIZE);
	    fprintf (stderr, "Adjust datatypes.h & rebuild\n");
	    exit (1);
	    }
	}	/* Each block */

/*
OBSOLETE - Takes a lot of space up, and isn't particularly useful
    fprintf (Inst_Logfile, " - %ld\n", BlockSequenceNum - 1);
*/
    ProcSequenceNum++;
    if (ProcSequenceNum > PROC_TABLE_TYPE_SIZE)
	{
	fprintf (stderr, "ProcSequenceNum (0x%ld) > PROC_TABLE_TYPE_SIZE (0x%ld)\n",
		ProcSequenceNum, PROC_TABLE_TYPE_SIZE);
	fprintf (stderr, "Adjust datatypes.h & rebuild\n");
	exit (1);
	}
    }	/* Each proc */
}

void InstrumentFini()
{
  long	i;
  long	netval;

  AddCallProto ("DynAllocateStorage(long, long)");
  AddCallProgram (ProgramBefore, "DynAllocateStorage",
			ProcSequenceNum, BlockSequenceNum);
  AddCallProto ("Dyn_ProcCountInit(long,char *,char *,long)");

  for (i = 1; i < ProcSequenceNum; i++)
    {
    AddCallProgram (ProgramBefore, "Dyn_ProcCountInit",
			i,
			ProcTable[i].ProcName,
			ProcTable[i].ProcFileName,
			ProcTable[i].StartAddr);
    }

  for (i = 1; i < BlockSequenceNum; i++)
    {
    AddCallProgram (ProgramBefore, "BbInfo",
			i,
			BlockTable[i].ProcId,
			BlockTable[i].Num_Insns,
			BlockTable[i].Block_Addr,
			BlockTable[i].StartLineNum,
			BlockTable[i].EndLineNum);
    }

  AddCallProto ("DynamicFini()");
  AddCallProgram (ProgramAfter, "DynamicFini");

  fprintf(Inst_Logfile, "A total of %d procedures found\n", ProcSequenceNum);
  fprintf(Inst_Logfile, "A total of %d blocks found\n", BlockSequenceNum);
  fprintf (Inst_Logfile, "*** Dynamic data begin ***\n");
  fprintf (Inst_Logfile, "NOTE: an 'Epoch' is the dynamic execution block number.\n");
  fprintf (Inst_Logfile, "Thus, Epoch 51 is the 51st block of code executed.\n");
  fclose(Inst_Logfile);

}

