
/*
 *	Copyright (C) 2000, Rick Gorton
 *	Written by Rick Gorton, July, 2000
 *	gorton@world.std.com
 *	All rights reserved
 *
 *	Helper functions for atom-izing time data collection.
 *	Names/capabilities sort of based upon those in FreePort Express
 *
 *	Currently implemented functions are:
 *		PrevIntReg		Prior instruction use or set
 *					of an integer register
 *		PrevIntRegUse		Prior use
 *		PrevIntRegSet		Prior set
 *		NextIntReg		Next use or set
 *		NextIntRegUse		Next use
 *		NextIntRegSet		Next set
 *
 *		PrevFloatReg		Prior instruction use or set
 *					of a float register
 *		PrevFloatRegUse		Prior use
 *		PrevFloatRegSet		Prior set
 *		NextFloatReg		Next use or set
 *		NextFloatRegUse		Next use
 *		NextFloatRegSet		Next set
 *
 *		SetsIntReg		Boolean: instruction sets the
 *					integer register
 *		UsesIntReg		Boolean: uses integer register
 *
 *		SetsFloatReg		Boolean: instruction sets the
 *					float register
 *		UsesFloatReg		Boolean: uses float register
 *
 *		Opcodes			Returns opcode enum of instruction
 *		GetLitValue		Returns literal field (0-255) of instruction
 *
 *
 * NOTE:
 *	If you find this tool useful, the author is interested in knowing
 *	how you used it, and what kind of modifications you found useful.
 *
 */
#include <stdio.h>
#include <cmplrs/atom.inst.h>
#include "datatypes.h"

/*
 * PrevIntReg - return NULL or pointer to relevant instruction
 */
Inst *PrevIntReg(Inst *this_inst, RegvType this_reg)
{
  Inst*            temp_inst;
  InstRegUsageVec  this_vec;
  long              mask;

  if (this_inst == NULL)
    return NULL;
  temp_inst = this_inst;
  while ((temp_inst = GetPrevInst(temp_inst)) != NULL)
    {
      GetInstRegUsage (temp_inst, &this_vec);   
      mask = 1 << (int) this_reg;
      if ((this_vec.ureg_bitvec[0] & mask) != 0)
        return temp_inst;	/* For a use of the register */
      if ((this_vec.dreg_bitvec[0] & mask) != 0)
        return temp_inst;	/* For a set of the register */
    } 
  return NULL; 
}

/*
/*
 * PrevIntRegUse - return NULL or pointer to relevant instruction
 */
Inst *PrevIntRegUse(Inst *this_inst, RegvType this_reg)
{
  Inst*            temp_inst;
  InstRegUsageVec  this_vec;
  long              mask;

  if (this_inst == NULL)
    return NULL;
  temp_inst = this_inst;
  while ((temp_inst = GetPrevInst(temp_inst)) != NULL)
    {
      GetInstRegUsage (temp_inst, &this_vec);   
      mask = 1 << (int) this_reg;
      if ((this_vec.ureg_bitvec[0] & mask) != 0)
        return temp_inst;
    } 
  return NULL; 
}

/*
 * PrevIntRegSet - return NULL or pointer to relevant instruction
 */
Inst *PrevIntRegSet(Inst *this_inst, RegvType this_reg)
{
  Inst*            temp_inst;
  InstRegUsageVec  this_vec;
  long              mask;

  if (this_inst == NULL)
    return NULL;
  temp_inst = this_inst;
  while ((temp_inst = GetPrevInst(temp_inst)) != NULL)
    {
      GetInstRegUsage (temp_inst, &this_vec);   
      mask = 1 << (int) this_reg;
      if ((this_vec.dreg_bitvec[0] & mask) != 0)
        return temp_inst;
    } 
  return NULL; 
}

/*
 * NextIntReg - return NULL or pointer to relevant instruction
 */
Inst *NextIntReg(Inst *this_inst, RegvType this_reg)
{
  Inst*            temp_inst;
  InstRegUsageVec  this_vec;
  long              mask;

  if (this_inst == NULL)
    return NULL;
  temp_inst = this_inst;
  while ((temp_inst = GetNextInst(temp_inst)) != NULL)
    {
      GetInstRegUsage (temp_inst, &this_vec);   
      mask = 1 << (int) this_reg;
      if ((this_vec.ureg_bitvec[0] & mask) != 0)
        return temp_inst;	/* For a use of the register */
      if ((this_vec.dreg_bitvec[0] & mask) != 0)
        return temp_inst;	/* For a set of the register */
    } 
  return NULL; 
}

/*
/*
 * NextIntRegUse - return NULL or pointer to relevant instruction
 */
Inst *NextIntRegUse(Inst *this_inst, RegvType this_reg)
{
  Inst*            temp_inst;
  InstRegUsageVec  this_vec;
  long              mask;

  if (this_inst == NULL)
    return NULL;
  temp_inst = this_inst;
  while ((temp_inst = GetNextInst(temp_inst)) != NULL)
    {
      GetInstRegUsage (temp_inst, &this_vec);   
      mask = 1 << (int) this_reg;
      if ((this_vec.ureg_bitvec[0] & mask) != 0)
        return temp_inst;
    } 
  return NULL; 
}

/*
 * NextIntRegSet - return NULL or pointer to relevant instruction
 */

Inst *NextIntRegSet(Inst *this_inst, RegvType this_reg)
{
  Inst*            temp_inst;
  InstRegUsageVec  this_vec;
  long              mask;

  if (this_inst == NULL)
    return NULL;
  temp_inst = this_inst;
  while ((temp_inst = GetNextInst(temp_inst)) != NULL)
    {
      GetInstRegUsage (temp_inst, &this_vec);   
      mask = 1 << (int) this_reg;
      if ((this_vec.dreg_bitvec[0] & mask) != 0)
        return temp_inst;
    } 
  return NULL; 
}

/*
 * PrevFloatReg - return NULL or pointer to relevant instruction
 */
Inst *PrevFloatReg(Inst *this_inst, RegvType this_reg)
{
  Inst*            temp_inst;
  InstRegUsageVec  this_vec;
  long              mask;

  if (this_inst == NULL)
    return NULL;
  temp_inst = this_inst;
  while ((temp_inst = GetPrevInst(temp_inst)) != NULL)
    {
      GetInstRegUsage (temp_inst, &this_vec);   
      mask = 1 << (int) this_reg;
      if ((this_vec.ureg_bitvec[0] & mask) != 0)
        return temp_inst;	/* For a set of the register */
      if ((this_vec.dreg_bitvec[0] & mask) != 0)
        return temp_inst;	/* For a set of the register */
    } 
  return NULL; 
}

/*
/*
 * PrevFloatRegUse - return NULL or pointer to relevant instruction
 */
Inst *PrevFloatRegUse(Inst *this_inst, RegvType this_reg)
{
  Inst*            temp_inst;
  InstRegUsageVec  this_vec;
  long              mask;

  if (this_inst == NULL)
    return NULL;
  temp_inst = this_inst;
  while ((temp_inst = GetPrevInst(temp_inst)) != NULL)
    {
      GetInstRegUsage (temp_inst, &this_vec);   
      mask = 1 << (int) this_reg;
      if ((this_vec.ureg_bitvec[0] & mask) != 0)
        return temp_inst;
    } 
  return NULL; 
}

/*
/*
 * PrevFloatRegSet - return NULL or pointer to relevant instruction
 */
Inst *PrevFloatRegSet(Inst *this_inst, RegvType this_reg)
{
  Inst*            temp_inst;
  InstRegUsageVec  this_vec;
  long              mask;

  if (this_inst == NULL)
    return NULL;
  temp_inst = this_inst;
  while ((temp_inst = GetPrevInst(temp_inst)) != NULL)
    {
      GetInstRegUsage (temp_inst, &this_vec);   
      mask = 1 << (int) this_reg;
      if ((this_vec.dreg_bitvec[0] & mask) != 0)
        return temp_inst;
    } 
  return NULL; 
}

/*
 * NextFloatReg - return NULL or pointer to relevant instruction
 */
Inst *NextFloatReg(Inst *this_inst, RegvType this_reg)
{
  Inst*            temp_inst;
  InstRegUsageVec  this_vec;
  long              mask;

  if (this_inst == NULL)
    return NULL;
  temp_inst = this_inst;
  while ((temp_inst = GetNextInst(temp_inst)) != NULL)
    {
      GetInstRegUsage (temp_inst, &this_vec);   
      mask = 1 << (int) this_reg;
      if ((this_vec.ureg_bitvec[0] & mask) != 0)
        return temp_inst;	/* For a use of the register */
      if ((this_vec.dreg_bitvec[0] & mask) != 0)
        return temp_inst;	/* For a set of the register */
    } 
  return NULL; 
}

/*
 * NextFloatRegUse - return NULL or pointer to relevant instruction
 */
Inst *NextFloatRegUse(Inst *this_inst, RegvType this_reg)
{
  Inst*            temp_inst;
  InstRegUsageVec  this_vec;
  long              mask;

  if (this_inst == NULL)
    return NULL;
  temp_inst = this_inst;
  while ((temp_inst = GetNextInst(temp_inst)) != NULL)
    {
      GetInstRegUsage (temp_inst, &this_vec);   
      mask = 1 << (int) this_reg;
      if ((this_vec.ureg_bitvec[0] & mask) != 0)
        return temp_inst;
    } 
  return NULL; 
}

/*
 * NextFloatRegSet - return NULL or pointer to relevant instruction
 */
Inst *NextFloatRegSet(Inst *this_inst, RegvType this_reg)
{
  Inst*            temp_inst;
  InstRegUsageVec  this_vec;
  long              mask;

  if (this_inst == NULL)
    return NULL;
  temp_inst = this_inst;
  while ((temp_inst = GetNextInst(temp_inst)) != NULL)
    {
      GetInstRegUsage (temp_inst, &this_vec);   
      mask = 1 << (int) this_reg;
      if ((this_vec.dreg_bitvec[0] & mask) != 0)
        return temp_inst;
    } 
  return NULL; 
}

BOOLEAN SetsIntReg(Inst *this_inst, RegvType this_reg)
{
  InstRegUsageVec	this_vec;
  unsigned long		mask;

  GetInstRegUsage (this_inst, &this_vec);
	/* Convert register number into a bit mask */
  mask = 1 << (int) this_reg;

  if ((this_vec.dreg_bitvec[0] & mask) != 0)
    return TRUE;
  return FALSE;
}

BOOLEAN UsesIntReg(Inst *this_inst, RegvType this_reg)
{
  InstRegUsageVec	this_vec;
  unsigned long		mask;

  GetInstRegUsage (this_inst, &this_vec);
	/* Convert register number into a bit mask */
  mask = 1 << (int) this_reg;

  if ((this_vec.ureg_bitvec[0] & mask) != 0)
    return TRUE;
  return FALSE;
}

BOOLEAN SetsFloatReg(Inst *this_inst, RegvType this_reg)
{
  InstRegUsageVec	this_vec;
  unsigned long		mask;

  GetInstRegUsage (this_inst, &this_vec);
	/* Convert register number into a bit mask */
  mask = 1 << (int) this_reg;

  if ((this_vec.dreg_bitvec[0] & mask) != 0)
    return TRUE;
  return FALSE;
}

BOOLEAN UsesFloatReg(Inst *this_inst, RegvType this_reg)
{
  InstRegUsageVec	this_vec;
  unsigned long		mask;

  GetInstRegUsage (this_inst, &this_vec);
	/* Convert register number into a bit mask */
  mask = 1 << (int) this_reg;

  if ((this_vec.ureg_bitvec[0] & mask) != 0)
    return TRUE;
  return FALSE;
}

int GetLitValue(Inst *this_inst)
{
  int	binhex;

  binhex = GetInstInfo(this_inst, InstBinary);
  return (LITERAL_VALUE_(binhex));
}
